class Analog(object):

    def __init__(self, parent):
        self._parent = parent
        
    '''
    ##############################################################################################################################
    Sensors
    ##############################################################################################################################
    '''
    def read_sensor(self, sensorID=0, NAvg=100):
        path = "/readanalogin_act"
        payload = {
            "readanaloginID": sensorID,
            "nanaloginavg": NAvg,
        }
        r = self._parent.post_json(path, payload)
        try:
            sensorValue = r['analoginVAL']
        except:
            sensorValue = None
        return sensorValue
    # TODO: Get/SET methods missing
    '''
    (*WifiController::getJDoc())["analoginVAL"] = returnValue;
    (*WifiController::getJDoc())["readanaloginPIN"] = analoginpin;
    (*WifiController::getJDoc())["N_analogin_avg"] = N_analogin_avg;
    (*WifiController::getJDoc())["readanaloginID"] = readanaloginID;
    (*WifiController::getJDoc())["N_analogin_avg"] = N_analogin_avg;
    '''

    def get_analog(self, readanaloginID=1):
        # readanaloginID 1,2,3
        path = "/readanalogin_get"
        payload = {
            "task": path,
            "readanaloginID": readanaloginID, 
            }
        r = self._parent.post_json(path, payload)
        return r
    
    def set_analog(self, readanaloginID=1, readanaloginPIN=15, nanaloginavg=1):
        # readanaloginID 1,2,3
        path = "/readanalogin_set"
        payload = {
            "task": path,
            "readanaloginID": readanaloginID, 
            "readanaloginPIN": readanaloginPIN, 
            "nanaloginavg": nanaloginavg
            }
        r = self._parent.post_json(path, payload)
        return r




#!/usr/bin/env python
# coding: utf-8
#%%
"""
Simple client code for the ESP32 in Python
Copyright 2021 Benedict Diederich, released under LGPL 3.0 or later
"""
from .mserial import Serial
try:
    from imswitch.imcommon.model import initLogger
    IS_IMSWITCH = True
except:
    print("No imswitch available")
    from .logger import Logger
    IS_IMSWITCH = False

import requests

class UC2Client(object):
    # headers = {'ESP32-version': '*'}
    headers={"Content-Type":"application/json"}
    getmessage = ""
    is_connected = False

    is_wifi = False
    is_serial = False

    # BAUDRATE = 500000
    BAUDRATE = 115200

    def __init__(self, host=None, port=31950, serialport=None, identity="UC2_Feather", baudrate=BAUDRATE, NLeds=64, DEBUG=False):
        '''
        This client connects to the UC2-REST microcontroller that can be found here
        https://github.com/openUC2/UC2-REST

        generally speaking you send/receive JSON documents that will cause an:
        1. action => "/XXX_act"
        2. getting => "/XXX_get"
        3. setting => "/XXX_set"

        you can send commands through wifi/http or usb/serial
        '''

        if IS_IMSWITCH:
            self.logger = initLogger(self, tryInheritParent=True)
        else:
            from .logger import Logger
            self.logger = Logger()
        # set default APIVersion
        self.APIVersion = 2

        # initialize communication channel (# connect to wifi or usb)
        if serialport is not None:
            # use USB connection
            self.serial = Serial(serialport, baudrate, parent=self, identity=identity, DEBUG=DEBUG)
            self.is_serial = True
            self.is_connected = True
            self.serial.DEBUG = DEBUG
        elif host is not None:
            # use client in wireless mode
            self.is_wifi = True
            self.host = host
            self.port = port

            # check if host is up
            self.logger.debug(f"Connecting to microscope {self.host}:{self.port}")
            self.is_connected = self.isConnected()
        else:
            self.logger.error("No ESP32 device is connected - check IP or Serial port!")

        # import libraries depending on API version
        if self.APIVersion == 1:
            self.logger.debug("Using API version 1")
            from .v1.galvo import Galvo
            from .v1.config import config
            from .v1.logger import Logger
            from .v1.ledmatrix import LedMatrix
            from .v1.motor import Motor
            from .v1.state import State
            from .v1.laser import Laser
            from .v1.wifi import Wifi
            from .v1.camera import Camera
            from .v1.analog import Analog
        elif self.APIVersion == 2:    
            self.logger.debug("Using API version 2")        
            from .galvo import Galvo
            from .config import config
            from .ledmatrix import LedMatrix
            from .motor import Motor
            from .home import Home
            from .state import State
            from .laser import Laser
            from .wifi import Wifi
            from .camera import Camera
            from .analog import Analog
            from .updater import updater
            from .modules import Modules
            from .digitalout import DigitalOut


        #FIXME
        #self.set_state(debug=False)

        # initialize state
        self.state = State(self)
        self.state.get_state()

        # initialize config
        self.config = config(self)

        # initialize LED matrix
        self.led = LedMatrix(self, NLeds=NLeds)

        # initilize motor
        self.motor = Motor(self)
        
        # initiliaze homing
        self.home = Home(self)

        # initialize laser
        self.state = State(self)

        # initialize galvo
        self.galvo = Galvo(self)

        # initialize laser
        self.laser = Laser(self)

        # initialize wifi
        self.wifi = Wifi(self)

        # initialize camera
        self.camera = Camera(self)

        # initialize analog
        self.analog = Analog(self)
        
        # initialize digital out
        self.digitalout = DigitalOut(self)
        
        # initialize config
        self.config = config(self)
        self.pinConfig = self.config.loadConfigDevice()
        
        # initialize updater 
        self.updater = updater(parent=self)
        
        # initialize module controller
        self.modules = Modules(parent=self)
   
    def post_json(self, path, payload, getReturn=True, timeout=1):
        if self.is_wifi:
            # FIXME: this is not working
            url = f"http://{self.host}:{self.port}{path}"
            r = requests.post(url, json=payload, headers=self.headers)
            return r.json()
        elif self.is_serial:
            return self.serial.post_json(path, payload, getReturn=getReturn, timeout=timeout)
        else:
            self.logger.error("No ESP32 device is connected - check IP or Serial port!")
            return None

    def get_json(self, path, timeout=1):
        if self.is_wifi:
            # FIXME: this is not working
            url = f"http://{self.host}:{self.port}{path}"
            r = requests.get(url, headers=self.headers)
            return r.json()
        elif self.is_serial:
            self.serial.get_json(path)
            return self.serial.read_json()
        else:
            self.logger.error("No ESP32 device is connected - check IP or Serial port!")
            return None



class Laser(object):
    ## Laser
    def __init__(self, parent):
        self._parent = parent
        self.filter_pos_1 = 0
        self.filter_pos_2 = 0
        self.filter_pos_3 = 0
        self.filter_pos_LED = 0
        self._parent.logger.debug("Attention, lasers are on channels 1,2,3")
        
    def set_laser(self, channel=1, value=0, auto_filterswitch=False,
                        filter_axis=-1, filter_position = None,
                        despeckleAmplitude = 0.,
                        despecklePeriod=10, timeout=20, is_blocking = True):
        if channel not in (0,1,2,3):
            if channel=="R":
                channel = 1
            elif channel=="G":
                channel = 2
            elif channel=="B":
                channel = 3

        if auto_filterswitch and value >0:
            if filter_position is None:
                if channel==1:
                    filter_position_toGo = self.filter_pos_1
                if channel==2:
                    filter_position_toGo = self.filter_pos_2
                if channel==3:
                    filter_position_toGo = self.filter_pos_3
                if channel=="LED":
                    filter_position_toGo = self.filter_pos_LED
            else:
                filter_position_toGo = filter_position

            self.switch_filter(filter_pos=filter_position_toGo, filter_axis=filter_axis, timeout=timeout,is_blocking=is_blocking)

        path = '/laser_act'
        
        payload = {
            "task": path,
            "LASERid": channel,
            "LASERval": value,
            "LASERdespeckle": int(value*despeckleAmplitude),
            "LASERdespecklePeriod": int(despecklePeriod),

        }
        self._parent.logger.debug("Setting Laser "+str(channel)+", value: "+str(value))
        r = self._parent.post_json(path, payload, getReturn=is_blocking)
        return r

    def set_laserpin(self, laserid=1, laserpin=0):
        path = '/laser_set'
        
        payload = {
            "task": path,
            "LASERid": laserid,
            "LASERpin": laserpin
        }
        
        r = self._parent.post_json(path, payload)
        return r

    def get_laserpins(self):
        path = '/laser_get'
        
        payload = {
            "task": path,
        }
        
        r = self._parent.post_json(path, payload)
        
        if type(r) is dict:
            # cast laser pins
            if "LASER1pin" in r: r["LASER1pin"] = int(r["LASER1pin"])
            else: r["LASER1pin"] = 0
            if r.__contains__("LASER2pin"): r["LASER2pin"] = int(r["LASER2pin"])
            else: r["LASER2pin"] = 0
            if r.__contains__("LASER3pin"): r["LASER3pin"] = int(r["LASER3pin"])
            else: r["LASER3pin"] = 0
        else:
            r={}
            r["LASER1pin"] = 0
            r["LASER2pin"] = 0
            r["LASER3pin"] = 0
        
        return r
    
    def get_laserpin(self, laserid=1):
        path = '/laser_get'
        
        payload = {
            "task": path,
        }
        
        r = self._parent.post_json(path, payload,timeout=2)
        return r["LASER"+str(laserid)+"pin"]



"""
__version__.py
~~~~~~~~~~~~~~

"""

__title__ = 'UC2 REST'
__description__ = 'This pacage will help you to drive the ESP32-driven microscopy control modules from UC2'
__version__ = "v0.2.0.8"
__author__ = 'Benedict Diederich'
__author_email__ = 'benedictdied@gmail.com'
__license__ = 'GPL v3'
__url__ = 'https://github.com/openUC2/UC2-REST'



import numpy as np
import time
import json


gTIMEOUT = 10 # seconds to wait for a response from the ESP32
class Motor(object):

    # indicate if there is any motion happening
    isRunning = False

    # a dictionary that stores all motor parameters for each dxis
    settingsdict = {"motor": {"steppers":
        [{"stepperid":0,"dir":0,"step":0,"enable":0,"dir_inverted":False,"step_inverted":False,"enable_inverted":False,"speed":0,"speedmax":200000,"max_pos":100000,"min_pos":-100000},
         {"stepperid":1,"dir":0,"step":0,"enable":0,"dir_inverted":False,"step_inverted":False,"enable_inverted":False,"speed":0,"speedmax":200000,"max_pos":100000,"min_pos":-100000},
         {"stepperid":2,"dir":0,"step":0,"enable":0,"dir_inverted":False,"step_inverted":False,"enable_inverted":False,"speed":0,"speedmax":200000,"max_pos":100000,"min_pos":-100000},
         {"stepperid":3,"dir":0,"step":0,"enable":0,"dir_inverted":False,"step_inverted":False,"enable_inverted":False,"speed":0,"speedmax":200000,"max_pos":100000,"min_pos":-100000}]
        }}



    def __init__(self, parent=None):
        self._parent = parent

        self.nMotors = 4
        self.steps_last = np.zeros((self.nMotors))
        self.backlash = np.zeros((self.nMotors))
        self.stepSize = np.ones((self.nMotors))
        self.maxStep = np.ones((self.nMotors))*np.inf
        self.minStep = np.ones((self.nMotors))*(-np.inf)

        self.minPosX = -np.inf
        self.minPosY = -np.inf
        self.minPosZ = -np.inf
        self.minPosT = -np.inf
        self.maxPosX = np.inf
        self.maxPosY = np.inf
        self.maxPosZ = np.inf
        self.maxPosT = np.inf
        self.stepSizeX =  1
        self.stepSizeY =  1
        self.stepSizeZ =  1
        self.stepSizeT =  1

        self.motorAxisOrder = [0,1,2,3] # motor axis is 1,2,3,0 => X,Y,Z,T # FIXME: Hardcoded

    def setMotorAxisOrder(self, order=[0,1,2,3]):
        self.motorAxisOrder = order

    '''################################################################################################################################################
    HIGH-LEVEL Functions that rely on basic REST-API functions
    ################################################################################################################################################'''
    def setup_motor(self, axis, minPos, maxPos, stepSize, backlash):
        if axis == "X":
            self.minPosX = minPos
            self.maxPosX = maxPos
            self.stepSizeX = stepSize
            self.backlashX = backlash
        elif axis == "Y":
            self.minPosY = minPos
            self.maxPosY = maxPos
            self.stepSizeY = stepSize
            self.backlashY = backlash
        elif axis == "Z":
            self.minPosZ = minPos
            self.maxPosZ = maxPos
            self.stepSizeZ = stepSize
            self.backlashZ = backlash
        elif axis == "T":
            self.minPosT = minPos
            self.maxPosT = maxPos
            self.stepSizeT = stepSize
            self.backlashT = backlash

    def xyztTo1230(self, axis):
        axis = axis.upper()
        if axis == "X":
            axis = 1
        if axis == "Y":
            axis = 2
        if axis == "Z":
            axis = 3
        if axis == "T" or axis == "A":
            axis = 0
        return axis

    def move_x(self, steps=100, speed=1000, is_blocking=False, is_absolute=False, is_enabled=True, timeout=gTIMEOUT):
        return self.move_axis_by_name(axis="X", steps=steps, speed=speed, is_blocking=is_blocking, is_absolute=is_absolute, is_enabled=is_enabled, timeout=timeout)

    def move_y(self, steps=100, speed=1000, is_blocking=False, is_absolute=False, is_enabled=True, timeout=gTIMEOUT):
        return self.move_axis_by_name(axis="Y", steps=steps, speed=speed, is_blocking=is_blocking, is_absolute=is_absolute, is_enabled=is_enabled, timeout=timeout)

    def move_z(self, steps=100, speed=1000, is_blocking=False, is_absolute=False, is_enabled=True, timeout=gTIMEOUT):
        return self.move_axis_by_name(axis="Z", steps=steps, speed=speed, is_blocking=is_blocking, is_absolute=is_absolute, is_enabled=is_enabled, timeout=timeout)

    def move_t(self, steps=100, speed=1000, is_blocking=False, is_absolute=False, is_enabled=True, timeout=gTIMEOUT):
        return self.move_axis_by_name(axis="T", steps=steps, speed=speed, is_blocking=is_blocking, is_absolute=is_absolute, is_enabled=is_enabled, timeout=timeout)

    def move_xyz(self, steps=(0,0,0), speed=(1000,1000,1000), is_blocking=False, is_absolute=False, is_enabled=True, timeout=gTIMEOUT):
        if len(speed)!= 3:
            speed = (speed,speed,speed)

        # motor axis is 1,2,3,0 => X,Y,Z,T # FIXME: Hardcoded
        r = self.move_xyzt(steps=(0,steps[0],steps[1],steps[2]), speed=(0,speed[0],speed[1],speed[2]), is_blocking=is_blocking, is_absolute=is_absolute, is_enabled=is_enabled, timeout=timeout)
        return r

    def move_xy(self, steps=(0,0), speed=(1000,1000), is_blocking=False, is_absolute=False, is_enabled=True, timeout=gTIMEOUT):
        if len(speed)!= 2:
            speed = (speed,speed)

        # motor axis is 1,2,3,0 => X,Y,Z,T # FIXME: Hardcoded
        r = self.move_xyzt(steps=(0, steps[0],steps[1],0), speed=(0,speed[0],speed[1],0), is_blocking=is_blocking, is_absolute=is_absolute, is_enabled=is_enabled, timeout=timeout)
        return r

    def move_xyzt(self, steps=(0,0,0,0), speed=(1000,1000,1000,1000), is_blocking=False, is_absolute=False, is_enabled=True, timeout=gTIMEOUT):
        if type(speed)==int:
            speed = (speed,speed,speed,speed)
        if type(steps)==int:
            steps = (steps,steps,steps,steps)

        r = self.move_stepper(steps=steps, speed=speed, backlash=(self.backlash[0],self.backlash[1],self.backlash[2],self.backlash[3]), is_blocking=is_blocking, is_absolute=is_absolute, is_enabled=is_enabled, timeout=timeout)
        return r

    def move_axis_by_name(self, axis="X", steps=100, speed=1000, is_blocking=False, is_absolute=False, is_enabled=True, timeout=gTIMEOUT):
        axis = self.xyztTo1230(axis)
        _speed=np.zeros(4)
        _speed[axis] = speed
        _steps=np.array((0,0,0,0))
        _steps[axis] = steps
        _backlash=np.zeros(4)
        _backlash[axis] = self.backlash[axis]
        r = self.move_stepper(_steps, speed=_speed, timeout=timeout, backlash=_backlash, is_blocking=is_blocking, is_absolute=is_absolute, is_enabled=is_enabled)
        return r

    def move_forever(self, speed=(0,0,0,0), is_stop=False):
        if type(speed)==int:
            speed=(speed, speed, speed, speed)
        if len(speed)==3:
            speed = (*speed,0)
        '''
        {"task":"/motor_act",
            "motor":
            {
                "steppers": [
                    { "stepperid": 3, "isforever": 1, "speed": 2000}
                ]
            }
        }
        '''
        # only consider those actions that are necessary
        motorPropList = []
        for iMotor in range(4):
            if abs(speed[iMotor])>0:
                motorProp = { "stepperid": iMotor,
                             "isforever": int(not is_stop),
                             "speed": speed[iMotor]}
                motorPropList.append(motorProp)

        path = "/motor_act"
        payload = {
            "task":path,
            "motor":
            {
                "steppers": motorPropList
            }
        }

        r = self._parent.post_json(path, payload, timeout=0)
        return r

    def stop(self, axis=None):
        axisNumberList = []
        if axis is None:
            for iMotor in range(self.nMotors):
                axisNumberList.append(iMotor)
        else:
            axisNumberList.append(self.xyztTo1230(axis))

        motorPropList = []
        for iMotor in axisNumberList:
            motorProp = { "stepperid": self.motorAxisOrder[iMotor],
                            "isstop": True}
            motorPropList.append(motorProp)

        path = "/motor_act"
        payload = {
            "task":path,
            "motor":
            {
                "steppers": motorPropList
            }
        }
        # ensure that nothing blocks this command!
        if self.isRunning:
            self._parent.serial.breakCurrentCommunication()
        r = self._parent.post_json(path, payload, getReturn=False, timeout=0)
        return r


    def move_stepper(self, steps=(0,0,0,0), speed=(1000,1000,1000,1000), is_absolute=(False,False,False,False), timeout=1, backlash=(0,0,0,0), is_blocking=True, is_enabled=True):
        '''
        This tells the motor to run at a given speed for a specific number of steps; Multiple motors can run simultaneously

        XYZT => 1,2,3,0
        '''

        # determine the axis to operate
        axisToMove = np.where(np.abs(speed)>0)

        if type(is_absolute)==bool:
            isAbsoluteArray = np.zeros((4))
            isAbsoluteArray[axisToMove] = is_absolute
        else:
            isAbsoluteArray = is_absolute
            
        # convert single elements to array
        if type(speed)!=list and type(speed)!=tuple and type(speed)!=np.ndarray:
            speed = np.array((speed,speed,speed,speed))

        # make sure value is an array
        if type(steps)==tuple:
            steps = np.array(steps)

        # detect change in directiongit config --global user.name "Your Name"
        for iMotor in range(4):
            if np.sign(self.steps_last[iMotor]) != np.sign(steps[iMotor]):
                # we want to overshoot a bit
                steps[iMotor] = steps[iMotor] + (np.sign(steps[iMotor])*backlash[iMotor])

        # get current position
        #_positions = self.get_position() # x,y,z,t = 1,2,3,0
        #pos_3, pos_0, pos_1, pos_2 = _positions[0],_positions[1],_positions[2],_positions[3]

        # convert to physical units
        steps[0] *= 1/self.stepSizeT
        steps[1] *= 1/self.stepSizeX
        steps[2] *= 1/self.stepSizeY
        steps[3] *= 1/self.stepSizeZ
        '''
        # check if within limits
        if pos_0+steps_0 > self.maxPosX or pos_0+steps_0 < self.minPosX:
            steps_0=0
        if pos_1+steps_1 > self.maxPosY or pos_1+steps_1 < self.minPosY:
            steps_1=0
        if pos_2+steps_2 > self.maxPosZ or pos_2+steps_2 < self.minPosZ:
            steps_2 = 0
        if pos_3+steps_3 > self.maxPosT or pos_3+steps_3 < self.minPosT:
            steps_3 = 0
        '''

        '''
        {"task":"/motor_act",
        "motor":{
            "steppers":[
                {
                    "stepperid":1,
                    "position":-1000,
                    "speed":2000,
                    "isabs":0,
                    "isaccel":0
                }
            ]
        }
        }

        {"task":"/motor_act",
    "motor":
    {
        "steppers": [
            { "stepperid": 1, "position": -100, "speed": 2000, "isabs": 0, "isaccel":0}
        ]
    }
}
        '''

        # only consider those actions that are necessary
        motorPropList = []
        for iMotor in range(4):
            if isAbsoluteArray[iMotor] or abs(steps[iMotor])>0:
                motorProp = { "stepperid": self.motorAxisOrder[iMotor],
                             "position": np.int(steps[iMotor]),
                             "speed": speed[iMotor],
                             "isabs": isAbsoluteArray[iMotor],
                             "isaccel":0}
                motorPropList.append(motorProp)

        path = "/motor_act"
        payload = {
            "task":path,
            "motor":
            {
                "steppers": motorPropList
            }
        }

        # safe steps to track direction for backlash compensatio
        for iMotor in range(self.nMotors):
            self.steps_last[iMotor] = steps[iMotor]

        # drive motor
        self.isRunning = True
        r = self._parent.post_json(path, payload, getReturn=is_blocking, timeout=timeout)

        # wait until job has been done        
        time0=time.time()
        steppersRunning = np.abs(np.array(steps))>0
        if is_blocking and self._parent.serial.is_connected:
            while True:
                time.sleep(0.05) # don'T overwhelm the CPU
                # see if already done
                try:
                    rMessage = self._parent.serial.serialdevice.readline().decode() # TODO: Make sure it's compatible with all motors running at the same time
                except Exception as e:
                    self._parent.logger.error(e)
                    rMessage = ""
                # check if message contains a motor that is done already
                if rMessage.find('isDone') >-1:
                    ''' TODO: This only checks for one motor!'''
                    try:
                        rMessage = rMessage.split("\r")[0].replace("'", '"')
                        mMessage = json.loads(rMessage)
                        for iElement in mMessage['steppers']:
                            if iElement['isDone']:
                                mNumber = self.motorAxisOrder[iElement['stepperid']]
                                steppersRunning[mNumber] = False
                    except:
                        pass
                    
                if np.sum(steppersRunning)==0:
                    break

                if time.time()-time0>timeout:
                    break


        # reset busy flag
        self.isRunning = False
        return r

    def isBusy(self, steps, timeout=1):
        path = "/motor_get"
        payload = {
            "task":path,
            "isbusy": 1
        }
        r = self._parent.post_json(path, payload, timeout=timeout)
        try:
            isbusy = 0
            for iMotor in range(self.nMotors):
                isbusy += r["motor"]["steppers"][iMotor]["isbusy"]
            if isbusy:
                return True
            else :
                return False
        except Exception as e:
            return False



    def set_motor(self, stepperid = 0, position = None, stepPin=None, dirPin=None, enablePin=None, maxPos=None, minPos=None, acceleration=None, isEnable=None, isBlocking=True, timeout=2):
        path = "/motor_set"
        payload = {"task":path,
                    "motor":{
                    "steppers": [
                        {   "stepperid": self.motorAxisOrder[stepperid]}]
                    }}

        if stepPin is not None: payload['motor']["steppers"][0]["step"] = stepPin
        if dirPin is not None: payload['motor']["steppers"][0]["dir"] = dirPin
        if enablePin is not None: payload['motor']["steppers"][0]["enable"] = enablePin
        if maxPos is not None: payload['motor']["steppers"][0]["max_pos"] = maxPos
        if minPos is not None: payload['motor']["steppers"][0]["min_pos"] = minPos
        if position is not None: payload['motor']["steppers"][0]["position"] = position
        if acceleration is not None: payload['motor']["steppers"][0]["acceleration"] = acceleration
        if isEnable is not None: payload['motor']["steppers"][0]["isen"] = isEnable

        # send command
        r = self._parent.post_json(path, payload, timeout=1)

        # wait until job has been done
        time0=time.time()
        if isBlocking:
            while self.isBusy((0,0,0,0)):
                time.sleep(0.1)
                if time.time()-time0>timeout:
                    break

        return r

    def set_motor_currentPosition(self, axis=0, currentPosition=10000):
        if type(axis)==str:
            axis = self.xyztTo1230(axis)

        r = self.set_motor(stepperid = axis, position = currentPosition)
        return r

    def set_motor_acceleration(self, axis=0, acceleration=10000):
        if type(axis)==str:
            axis = self.xyztTo1230(axis)

        r = self.set_motor(stepperid = axis, acceleration=acceleration)
        return r

    def set_motor_enable(self, axis =0, is_enable=1):
        if type(axis)==str:
            axis = self.xyztTo1230(axis)
        path = "/motor_act"
        payload = {
            "task": path,
            "isen": is_enable
        }
        r = self._parent.post_json(path, payload)
        return r

    def get_position(self, axis=None, timeout=1):
        # pulls all current positions from the stepper controller
        path = "/motor_get"
        payload = {
            "task":path,
            "position":True,
        }
        r = self._parent.post_json(path, payload, timeout=timeout)
        _position = np.array((0,0,0,0)) # T,X,Y,Z
        _physicalStepSizes = np.array((self.stepSizeT, self.stepSizeX, self.stepSizeY, self.stepSizeZ))
        try:
            for index, istepper in enumerate(r["motor"]["steppers"]):
                _position[istepper["stepperid"]]=istepper["position"]*_physicalStepSizes[self.motorAxisOrder[index]]
        except Exception as e: self._parent.logger.error(e)
        return _position

    def set_position(self, axis=1, position=0, timeout=1):
        self._parent.logger.error("set_position is deprecated. Use set_motor_currentPosition instead.")
        '''
        path = "/motor_set"
        if axis=="X": axis=1
        if axis=="Y": axis=2
        if axis=="Z": axis=3

        payload = {
            "task":path,
            "axis":axis,
            "currentposition": position
        }
        r = self._parent.post_json(path, payload, timeout=timeout)

        return r
        '''
        return False

    def set_motors(self, settingsdict, isBlocking=True, timeout=1):
        path = "/motor_set"
        payload = settingsdict
        payload["task"] = path

        # send command
        r = self._parent.post_json(path, payload, timeout=1)

        # wait until job has been done
        time0=time.time()
        if isBlocking:
            while self.isBusy((0,0,0,0)):
                time.sleep(0.1)
                if time.time()-time0>timeout:
                    break

        return r


    def get_motor(self, axis=1, timeout=1):
        path = "/motor_get"
        payload = {
            "task":path,
        }
        r = self._parent.post_json(path, payload, timeout=timeout)
        try: return r["steppers"][self.motorAxisOrder[axis]]
        except: return False

    def get_motors(self, timeout=1):
        path = "/motor_get"
        payload = {
            "task":path,
        }
        r = self._parent.post_json(path, payload, timeout=timeout)
        try: return r["steppers"]
        except: return False


    def set_direction(self, axis=1, sign=1, timeout=1):
        return False



import serial
import serial.tools.list_ports
import time
import json

T_SERIAL_WARMUP = 5 # the time to wait for the serial to warm up

class Serial(object):
    
    def __init__(self, port, baudrate, timeout=1, identity="UC2_Feather", parent=None, DEBUG=False):
        self.serialport = port
        self.baudrate = baudrate
        self.timeout = timeout
        self._parent = parent
        self.identity = identity
        self.DEBUG = DEBUG

        self.isSafetyBreak = False
        
        # default version (v1 or v2) for the API
        self.versionFirmware = "V2.0"
        
        self.NumberRetryReconnect = 0
        self.MaxNumberRetryReconnect = 0
        
        self.open() # creates self.serialdevice
        
        # switch between different api versions 
        if self.versionFirmware == "V1.2":
            self._parent.APIVersion = 1
        elif self.versionFirmware == "V2.0":
            self._parent.APIVersion = 2
        
        
    def open(self):
        '''Open the serial port'''
        self.is_connected = False
        self.serialdevice = None
        try:
            # most simple case: We know all parameters
            self.serialdevice = serial.Serial(port=self.serialport, baudrate=self.baudrate, timeout=1)
            self.is_connected = True
            time.sleep(T_SERIAL_WARMUP) # let it warm up
            correctFirmware = self.checkFirmware()
            if not correctFirmware:
                raise Exception("Wrong firmware")
                        
            self._parent.logger.debug("We are connected: "+str(self.is_connected) + " on port: "+self.serialdevice.port)
            return self.serialdevice
        except Exception as e:
            # try to find the PORT
            self._parent.logger.error(e)
            _available_ports = serial.tools.list_ports.comports(include_links=False)

            portslist = ("COM", "/dev/tt", "/dev/a", "/dev/cu.SLA","/dev/cu.wchusb", "/dev/cu.usbserial") # TODO: Hardcoded :/
            descriptionlist = ("CH340", "CP2102")
            for iport in _available_ports:
                # list of possible serial ports
                self._parent.logger.debug(iport.device)
                if iport.device.startswith(portslist) or iport.description.startswith(descriptionlist):
                    try:
                        self.serialdevice = serial.Serial(port=iport.device, baudrate=self.baudrate, timeout=1)
                        self.serialdevice.write_timeout=1
                        self.is_connected = True # attempting to initiliaze connection
                        time.sleep(T_SERIAL_WARMUP) # let it warm up and wait until debugging messages may vanish
                        correctFirmware = self.checkFirmware()
                        if correctFirmware:
                            self.serialport = iport.device
                            self._parent.logger.debug("We are connected: "+str(self.is_connected) + " on port: "+self.serialdevice.port)
                            self.NumberRetryReconnect=0
                            return self.serialdevice
                    except Exception as e:
                        self._parent.logger.debug("Trying out port "+iport.device+" failed")
                        self._parent.logger.error(e)
                        self.is_connected = False
        # last resort: we are not connected
        self.serialport = "NotConnected"
        self.serialdevice = SerialDeviceDummy()
        self._parent.logger.debug("No USB device connected! Using DUMMY!")
        return self.serialdevice                

    def checkFirmware(self, timeout=1):
        """Check if the firmware is correct"""
        path = "/state_get"
        _state = self.post_json(path, {"task":path}, timeout=timeout)
        try:
            self.versionFirmware = _state["identifier_id"]
        except Exception as e:
            self._parent.logger.error(e)
            self.versionFirmware = "V2.0"
            
        if _state["identifier_name"] == self.identity: 
            return True
        
        else: return False


    def closeSerial(self):
        self.serialdevice.close()
        
    def reconnect(self):
        """Reconnect to serial device"""
        if self.is_serial:
            self.initSerial(self.serialport, self.baudrate)

    def get_json(self, path):
        """Perform an HTTP GET request and return the JSON response"""
        message = {"task":path}
        message = json.dumps(message)
        self.serialdevice.flushInput()
        self.serialdevice.flushOutput()
        self._parent.logger.debug(message)
        returnmessage = self.serialdevice.write(message.encode(encoding='UTF-8'))
        return returnmessage
    
    def post_json(self, path, payload={}, getReturn=True, timeout=1):
        """Make an HTTP POST request and return the JSON response"""
        if "task" not in payload:
            payload["task"] = path
        
        if "isblock" in payload:
            is_blocking = payload["isblock"]
        else:
            is_blocking = True

        # write message to the serial
        self.writeSerial(payload)
        
        if getReturn:
            # we read the return message
            #self._parent.logger.debug(payload)
            returnmessage = self.readSerial(is_blocking=is_blocking, timeout=timeout)
            if returnmessage == 'deserializeJson() failed: NoMemory':
                # TODO: We will loose values here - need to set xyz coordinates!!
                self.serialdevice.close()
                self._parent.state.espRestart()
                self._parent.state.restart()
                self.open()
        else:
            returnmessage = False
        return returnmessage
        
    def writeSerial(self, payload):
        """Write JSON document to serial device"""
        try:
            if self.serialport == "NotConnected" and self.NumberRetryReconnect<self.MaxNumberRetryReconnect:
                # try to reconnect       
                self._parent.logger.debug("Trying to reconnect to serial device: "+str(self.NumberRetryReconnect))       
                self.NumberRetryReconnect += 1
                self.open()
                
            self.serialdevice.flushInput()
            self.serialdevice.flushOutput()
        except Exception as e:
            self._parent.logger.error(e)
            try:
                del self.serialdevice
            except:
                pass
            self.is_connected=False
            # attempt to reconnect?
            try:
                self.open()
            except:
                return -1

        if type(payload)==dict:
            payload = json.dumps(payload)
        try:
            if self.DEBUG: self._parent.logger.debug(payload)
            self.serialdevice.write(payload.encode(encoding='UTF-8'))
        except Exception as e:
            self._parent.logger.error(e)

    def breakCurrentCommunication(self):
        # this breaks the wait-for-return command in readserial
        self.isSafetyBreak = True
        self.serialdevice.flushInput()
        self.serialdevice.flushOutput()


    def readSerial(self, is_blocking=True, timeout = 1): # TODO: hardcoded timeout - not code
        """Receive and decode return message"""
        returnmessage = ''
        _returnmessage = ''
        rmessage = ''
        _time0 = time.time()
        if is_blocking:
            while is_blocking and not self.isSafetyBreak:
                try:
                    rmessage =  self.serialdevice.readline().decode()
                    if self.DEBUG: self._parent.logger.debug(rmessage)
                    returnmessage += rmessage
                    if rmessage.find("--")==0:
                        break
                except:
                    pass
                if (time.time()-_time0)>timeout:
                    break
            # casting to dict
            try:
                # TODO: check if this is a valid JSON
                _returnmessage = returnmessage.split("\n--")[0].split("\n++")[-1].replace("\r","").replace("\n", "").replace("'", '"')
                if self.DEBUG: self._parent.logger.debug(returnmessage)
                _returnmessage = json.loads(_returnmessage)
            except Exception as e:
                if self.DEBUG: self._parent.logger.debug("Casting json string from serial to Python dict failed")
            self.isSafetyBreak = False
        return _returnmessage
        
class SerialDummy(object):
        
    def __init__(self, port, baudrate, timeout=1, parent=None):
        self.port = port
        self.baudrate = baudrate
        self.timeout = timeout
        self._parent = parent
        
        self.serialdevice = self.open()
        
    def open(self):
        '''Open the serial port'''
        return SerialDeviceDummy()                

    def checkFirmware(self):
        """Check if the firmware is correct"""
        return True
        
    def close(self):
        self.closeSerial()
        
    def closeSerial(self):
        self.serialdevice.close()
        
    def reconnect(self):
        """Reconnect to serial device"""
        if self.is_serial:
            self.initSerial(self.serialport, self.baudrate)

    def get_json(self, path):
        """Perform an HTTP GET request and return the JSON response"""
        message = {"task":path}
        message = json.dumps(message)
        self.serialdevice.flushInput()
        self.serialdevice.flushOutput()
        returnmessage = self.serialdevice.write(message.encode(encoding='UTF-8'))
        return returnmessage

    def post_json(self, path, payload={}, getReturn=True, timeout=1):
        """Make an HTTP POST request and return the JSON response"""
        try:
            payload["task"]
        except:
            payload["task"] = path
        try:
            is_blocking = payload['isblock']
        except:
            is_blocking = True
        self.writeSerial(payload)
        if getReturn:
            # we read the return message
            #self._parent.logger.debug(payload)
            returnmessage = self.readSerial(is_blocking=False, timeout=timeout)
        else:
            returnmessage = None
        return returnmessage
        
    def writeSerial(self, payload):
        """Write JSON document to serial device"""
        try:
            self.serialdevice.flushInput()
            self.serialdevice.flushOutput()
        except Exception as e:
            self._parent.logger.error(e)
            try:
                del self.serialdevice
            except:
                pass
            self.is_connected=False
            # attempt to reconnect?
            try:
                self.open()
            except:
                return -1

        if type(payload)==dict:
            payload = json.dumps(payload)
        try:
            self.serialdevice.write(payload.encode(encoding='UTF-8'))
        except Exception as e:
            self._parent.logger.error(e)


    def readSerial(self, is_blocking=True, timeout = 15): # TODO: hardcoded timeout - not code
        """Receive and decode return message"""
        returnmessage = ''
        rmessage = ''
        _time0 = time.time()
        if is_blocking:
            while is_blocking:
                try:
                    rmessage =  self.serialdevice.readline().decode()
                    #self._parent.logger.debug(rmessage)
                    returnmessage += rmessage
                    if rmessage.find("--")==0:
                        break
                except:
                    pass
                if (time.time()-_time0)>timeout:
                    break
            # casting to dict
            try:
                returnmessage = json.loads(returnmessage.split("--")[0].split("++")[-1])
            except:
                self._parent.logger.debug("Casting json string from serial to Python dict failed")
                returnmessage = ""
        return returnmessage

        
class SerialDeviceDummy(object):
        def __init__(self) -> None:
            pass
        
        def close(self):
            pass
        
        def flushInput(self):
            pass
        
        def flushOutput(self):
            pass
        
        def write(self, payload):
            pass
        
        def readline(self):
            return b'{"task":"dummy"}--'


class Modules(object):
    ## Laser
    def __init__(self, parent):
        self._parent = parent
        # default settings
        self.default_modules = {
                "led":1,
                "motor":1,
                "slm":0,
                "home":1,
                "analogin":0,
                "pid":0,
                "laser":1,
                "dac":0,
                "analogout":0,
                "digitalout":0,
                "digitalin":1,
                "scanner":0,
                "wifi":0
            }
            
    def get_default_modules(self):
        return self.default_modules

    def get_modules(self, timeout=1):
        # returns a list of available modules that are set or not set in the ESP
        path = '/modules_get'
        
        payload = {
            "task": path,
            }
        r = self._parent.post_json(path, payload)
        
        return r

    def set_modules(self, modules, timeout=1):
        path = '/modules_set'
        
        payload = {
            "task": path,
            "modules": modules
        }
        
        r = self._parent.post_json(path, payload)
        return r




import json


class config(object):
    def __init__(self, parent, configFilePath=None): 
        self._parent = parent
        self.configFilePath = configFilePath
        
        if self.configFilePath is not None:
            self.configFilePath = "./config.json"
            #TODO: read file and load
            
        self.configFile = self.loadDefaultConfig()
    

    '''
    default Configuration
    '''     
        
    def loadDefaultConfig(self):
        self.defaultConfig = {"motorconfig": 
                [{"stepperid": 0, "dir": 18, "step": 19, "enable": 12, "dir_inverted": False, "step_inverted": False, "enable_inverted": False, "speed": 0, "speedmax": 200000, "max_pos": 100000, "min_pos": -100000}, 
                {"stepperid": 1, "dir": 16, "step": 26, "enable": 12, "dir_inverted": False, "step_inverted": False, "enable_inverted": False, "speed": 0, "speedmax": 200000, "max_pos": 100000, "min_pos": -100000}, 
                {"stepperid": 2, "dir": 27, "step": 25, "enable": 12, "dir_inverted": False, "step_inverted": False, "enable_inverted": False, "speed": 0, "speedmax": 200000, "max_pos": 100000, "min_pos": -100000}, 
                {"stepperid": 3, "dir": 14, "step": 17, "enable": 12, "dir_inverted": False, "step_inverted": False, "enable_inverted": False, "speed": 0, "speedmax": 200000, "max_pos": 100000, "min_pos": -100000}], 
            "ledconfig": {"ledArrNum": 0, "ledArrPin": 0, "LEDArrMode": [0, 1, 2, 3, 4, 5, 6, 7], "led_ison": False}, 
            "stateconfig": {"identifier_name": "UC2_Feather", "identifier_id": "V1.2", "identifier_date": "Nov  7 202212:52:14", "identifier_author": "BD", "IDENTIFIER_NAME": ""},
            "laserconfig": {"LASER1pin": 0, "LASER2pin": 0, "LASER3pin": 0}}            
        return self.defaultConfig
    
    def setDefaultConfig(self, configDict=None, configFile=None):
        if configDict is not None:
            self.defaultConfig = configDict
        if configFile is not None:
            self.defaultConfig = self.json2dict(configFile)
    
    
    def checkIfInitializedConfig(self, currentConfig, defaultConfig):
        ''' This should check if the different components were initialized with the config file
            Otherwise we won't be able to retrieve any signals from the device
        '''
        # TODO: Not implemented yet! 
        if currentConfig['motorconfig'] == defaultConfig['motorconfig']:
            print("motorconfig not initialized")
            return False

    def json2dict(self, jsonFile):
        with open(jsonFile) as f:
            data = json.load(f)
        return data
        
    def checkConfig(self, configfile):
        # check if config is valid
        if type(configfile['motorconfig'])==list and len(configfile['motorconfig'])==4:
            print("loaded config is valid")
            return True
        else:
            print("config is not valid")
            return False
        
        
    '''
    send configuration to device
    '''
    
    '''################################################################################################################################################
    Set Configurations
    ################################################################################################################################################'''

    def loadConfigDevice(self, timeout=1):
        # load current configuration from device 
        if 0:
            path = '/config_get'
            payload = {
            }
            r = self._parent.post_json(path, payload, timeout=timeout)
                        
            if type(r) != dict:
                r = self.loadDefaultConfig()
            else:
                self.setDefaultConfig(r)
            return r
        
        else:
            motorconfig = self._parent.motor.get_motors()
            ledconfig = self._parent.led.get_ledpin()
            laserconfig = self._parent.laser.get_laserpins()
            stateconfig = self._parent.state.get_state()
            self.configfile = {"motorconfig": motorconfig, 
                          "ledconfig": ledconfig, 
                          "laserconfig": laserconfig, 
                          "stateconfig": stateconfig}

            # check if config is valid
            isValidConfig = self.checkConfig(self.configfile)

            if isValidConfig:
                return self.configfile
            else:
                return self.loadDefaultConfig()

            

    def setConfigDevice(self, configfile, timeout=1):
        # push changes back to device
        if 0:
            path = '/config_set'
            if type(configfile)==dict:
                payload = configfile
            else: 
                return None
            r = self._parent.post_json(path, payload, timeout=timeout)
            return r
        else: 
            # set all lasers 
            notAllowedPins = (1,2,3,5)
            if configfile["laserconfig"]["LASER1pin"] in notAllowedPins:
                self._parent.laser.set_laserpin(1, configfile["laserconfig"]["LASER1pin"])
            if configfile["laserconfig"]["LASER2pin"] in notAllowedPins:
                self._parent.laser.set_laserpin(2, configfile["laserconfig"]["LASER2pin"])
            if configfile["laserconfig"]["LASER3pin"] in notAllowedPins:
                self._parent.laser.set_laserpin(3, configfile["laserconfig"]["LASER3pin"])
            self._parent.laser.set_laserpin(laserid=1, laserpin=configfile["laserconfig"]["LASER1pin"])
            self._parent.laser.set_laserpin(laserid=2, laserpin=configfile["laserconfig"]["LASER2pin"])
            self._parent.laser.set_laserpin(laserid=3, laserpin=configfile["laserconfig"]["LASER3pin"])

            # set led 
            if configfile["ledconfig"]["ledArrPin"] in notAllowedPins:
                self._parent.led.set_ledpin(ledArrPin=configfile["ledconfig"]["ledArrPin"], ledArrNum=configfile["ledconfig"]["ledArrNum"])
            self._parent.led.set_ledpin(ledArrPin=configfile["ledconfig"]["ledArrPin"], ledArrNum=configfile["ledconfig"]["ledArrNum"])
            
                
            # set motors
            if configfile["motorconfig"][0]["dir"] in notAllowedPins:
                configfile["motorconfig"][0]["dir"] = 0
            if configfile["motorconfig"][0]["step"] in notAllowedPins:
                configfile["motorconfig"][0]["step"] = 0
            if configfile["motorconfig"][1]["dir"] in notAllowedPins:
                configfile["motorconfig"][1]["dir"] = 0
            if configfile["motorconfig"][1]["step"] in notAllowedPins:
                configfile["motorconfig"][1]["step"] = 0
            if configfile["motorconfig"][2]["dir"] in notAllowedPins:
                configfile["motorconfig"][2]["dir"] = 0
            if configfile["motorconfig"][2]["step"] in notAllowedPins:
                configfile["motorconfig"][2]["step"] = 0
            if configfile["motorconfig"][3]["dir"] in notAllowedPins:
                configfile["motorconfig"][3]["dir"] = 0
            if configfile["motorconfig"][3]["step"] in notAllowedPins:
                configfile["motorconfig"][3]["step"] = 0   
            if configfile["motorconfig"][0]["enable"] in notAllowedPins:
                configfile["motorconfig"][0]["enable"] = 0
                configfile["motorconfig"][1]["enable"] = 0
                configfile["motorconfig"][2]["enable"] = 0
                configfile["motorconfig"][3]["enable"] = 0
                                                         
            motorconfig = {
                "motor":{
                    "steppers": 
                        configfile["motorconfig"]}}
            # check if pins are valid
            
            self._parent.motor.set_motors(settingsdict=motorconfig)
            




class Galvo(object):
    def __init__(self, parent=None):
        '''
        defaults:
            dac->Setup(DAC_CHANNEL_1, 0, 1, 0, 0, 2);
            dac->Setup(dac_channel, clk_div, frequency, scale, phase, invert);
      '''
        self.channel= 1
        self.frequency = 1
        self.offset = 0
        self.amplitude = amplitude=1/2
        self.clk_div = 0
        self.path = "/dac_act"
        
        self._parent = parent

    
    def set_dac(self, channel=1, frequency=1, offset=0, amplitude=1/2, clk_div=0, timeout=1):
        
        path = "/dac_act"
        payload = {
            "task": path,
            "dac_channel": channel, # 1 or 2
            "frequency": frequency,
            "offset": offset,
            "amplitude":amplitude,
            "clk_div": clk_div
        }
                        
        self._parent.post_json(path, payload, timeout=timeout, getReturn=False)

    '''
    ##############################################################################################################################
    SCANNER
    ##############################################################################################################################
    '''
    def set_scanner_pattern(self, numpyPattern, scannernFrames=1,
            scannerLaserVal=32000,
            scannerExposure=500, scannerDelay=500, is_blocking = False):

        scannerMode="pattern"
        path = '/scanner_act'
        arraySize = int(np.prod(numpyPattern.shape))
        payload = {
            "task":path,
            "scannernFrames":scannernFrames,
            "scannerMode":scannerMode,
            "arraySize":arraySize,
            "i":numpyPattern.flatten().tolist(),
            "scannerLaserVal":scannerLaserVal,
            "scannerExposure":scannerExposure,
            "scannerDelay":scannerDelay,
            "isblock": is_blocking
            }

        r = self.post_json(path, payload)
        return r

    def set_scanner_classic(self, scannernFrames=100,
            scannerXFrameMin=0, scannerXFrameMax=255,
            scannerYFrameMin=0, scannerYFrameMax=255,
            scannerEnable=0, scannerxMin=1,
            scannerxMax=5, scanneryMin=1,
            scanneryMax=5, scannerXStep=25,
            scannerYStep=25, scannerLaserVal=32000,
            scannerExposure=500, scannerDelay=500):

        scannerModec="classic",
        path = '/scanner_act'
        payload = {
            "task":path,
            "scannernFrames":scannernFrames,
            "scannerMode":scannerModec,
            "scannerXFrameMin":scannerXFrameMin,
            "scannerXFrameMax":scannerXFrameMax,
            "scannerYFrameMin":scannerYFrameMin,
            "scannerYFrameMax":scannerYFrameMax,
            "scannerEnable":scannerEnable,
            "scannerxMin":scannerxMin,
            "scannerxMax":scannerxMax,
            "scanneryMin":scanneryMin,
            "scanneryMax":scanneryMax,
            "scannerXStep":scannerXStep,
            "scannerYStep":scannerYStep,
            "scannerLaserVal":scannerLaserVal,
            "scannerExposure":scannerExposure,
            "scannerDelay":scannerDelay}

        r = self.post_json(path, payload)
        return r


    def set_galvo_freq(self, axis=1, value=1000):
        if axis+1 == 1:
            self.galvo1.frequency=value
            payload = self.galvo1.return_dict()
        else:
            self.galvo2.frequency=value
            payload = self.galvo2.return_dict()

        r = self.post_json(payload["task"], payload, timeout=1)
        return r

    def set_galvo_amp(self, axis=1, value=1000):
        if axis+1 == 1:
            self.galvo1.amplitude=value
            payload = self.galvo1.return_dict()
        else:
            self.galvo2.amplitude=value
            payload = self.galvo2.return_dict()

        r = self.post_json(payload["task"], payload, timeout=1)
        return r


class State(object):
    '''
    ##############################################################################################################################
    STATE
    ##############################################################################################################################
    '''

    def __init__(self, parent):
        self._parent = parent
        
    def get_state(self, timeout=3):
        path = "/state_get"

        payload = {
            "task":path
        }
        r = self._parent.post_json(path, payload, timeout=timeout)
        return r

    def set_state(self, debug=False, timeout=1):
        path = "/state_set"

        payload = {
            "task":path,
            "isdebug":int(debug)
        }
        r = self._parent.post_json(path, payload, timeout=timeout)
        return r

    def isControllerMode(self, timeout=1):
        # returns True if PS controller is active
        path = "/state_get"
        payload = {
            "task":path,
            "pscontroller": 1
        }
        r = self._parent.post_json(path, payload, timeout=timeout)
        try:
            return r["pscontroller"]
        except:
            return False

    def espRestart(self,timeout=1):
        # if isController =True=> only PS jjoystick will be accepted
        path = "/state_act"
        payload = {
            "restart":1
            }
        r = self._parent.post_json(path, payload, timeout=timeout)
        return r

    def setControllerMode(self, isController=False, timeout=1):
        # if isController =True=> only PS jjoystick will be accepted
        path = "/state_act"
        payload = {
            "task":path,
            "pscontroller": isController
        }
        r = self._parent.post_json(path, payload, timeout=timeout)
        return r

    def isBusy(self, timeout=1):
        path = "/state_get"
        payload = {
            "task":path,
            "isBusy": 1
        }
        r = self._parent.post_json(path, payload, timeout=timeout)
        try:
            return r["isBusy"]
        except:
            return r




    
class Wifi(object):
    
    def __init__(self, parent):
        self._parent = parent
        
    def isConnected(self):
        # check if client is connected to the same network
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            s.connect((self.host, int(self.port)))
            s.settimeout(2)
            s.shutdown(2)
            return True
        except:
            return False

    
    def get_json(self, path):
        if self.is_connected and self.is_wifi:
            if not path.startswith("http"):
                path = self.base_uri + path
            try:
                r = requests.get(path)
                r.raise_for_status()
                self.is_connected = True

                self.getmessage = r.json()
                self.is_sending = False
                return self.getmessage

            except Exception as e:
                if IS_IMSWITCH: self.__logger.error(e)
                self.is_connected = False
                self.is_sending = False
                # not connected
                return None

    def post_json(self, path, payload={}, headers=None, isInit=False, timeout=1):
        """Make an HTTP POST request and return the JSON response"""
        if self.is_connected and self.is_wifi:
            if not path.startswith("http"):
                path = self.base_uri + path
            if headers is None:
                headers = self.headers
            try:
                r = requests.post(path, json=payload, headers=headers,timeout=timeout)
                r.raise_for_status()
                r = r.json()
                self.is_connected = True
                self.is_sending = False
                return r
            except Exception as e:
                if IS_IMSWITCH: self.__logger.error(e)
                self.is_connected = False
                self.is_sending = False
                # not connected
                return None

 
    @property
    def base_uri(self):
        return f"http://{self.host}:{self.port}"


    def send_jpeg(self, image):
        if is_cv2:
            temp = NamedTemporaryFile()

            #add JPEG format to the NamedTemporaryFile
            iName = "".join([str(temp.name),".jpg"])

            #save the numpy array image onto the NamedTemporaryFile
            cv2.imwrite(iName,image)
            _, img_encoded = cv2.imencode('test.jpg', image)

            content_type = 'image/jpeg'
            headers = {'content-type': content_type}
            payload = img_encoded.tostring()
            path = '/uploadimage'

            #r = self.post_json(path, payload=payload, headers = headers)
            #requests.post(self.base_uri + path, data=img_encoded.tostring(), headers=headers)
            files = {'media': open(iName, 'rb')}
            if self.is_connected:
                requests.post(self.base_uri + path, files=files)



    def scanWifi(self, timeout=3):
        path = '/wifi/scan'
        payload={
            "task": path
        }
        r = self._parent.post_json(path, payload, timeout=timeout)
        return r

        


class DigitalOut(object):

    def __init__(self, parent):
        self._parent = parent
        self._logger = parent.logger
    
    
    def setup_digitaloutpin(self, id=1, pin=4):
        path = '/digitalout_set'
        payload = {
            "task": path,
            "digitaloutid": id,
            "digitaloutpin": pin,
        }
        r = self._parent.post_json(path, payload)
        return r
    
    def reset_triggertable(self):
        path = '/digitalout_act'
        payload = {
            "task": path,
            "digitaloutistriggerreset": 1,
        }
        r = self._parent.post_json(path, payload)
        return r
    
    def set_trigger(self, trigger1=False, delayOn1=0, delayOff1=0, trigger2=False, delayOn2=0, delayOff2=0, trigger3=False, delayOn3=0, delayOff3=0):
        '''
        this stats a trigger table with 3 triggers
        
        Parameters:
        
        '''
        path = '/digitalout_act'
        
        payload = {
            "task":path,
            "digitalout1TriggerDelayOn":delayOn1,
            "digitalout1TriggerDelayOff":delayOff1,
            "digitalout1IsTrigger":trigger1,
            "digitalout2TriggerDelayOn":delayOn2,
            "digitalout2TriggerDelayOff":delayOff2,
            "digitalout2IsTrigger":trigger2, 
            "digitalout3TriggerDelayOn":delayOn3,
            "digitalout3TriggerDelayOff":delayOff3,
            "digitalout3IsTrigger":trigger3,
        }
        
        r = self._parent.post_json(path, payload)
        return r
    
        


    def sendTrigger(self, triggerId=0):
        path = '/digital_act'

        payload = {
            "task": path,
            "digitalid": triggerId,
            "digitalval": -1,
        }

        r = self._parent.post_json(path, payload)
        return r



import numpy as np
import json
import time 
gTimeout = 0.1
class LedMatrix(object):
    def __init__(self, parent, NLeds=64):
        #TOOD: This is for the LED matrix only!
        self.NLeds = NLeds
        #assuming we have a square grid!
        self.Nx = self.Ny = int(np.sqrt(NLeds))
        
        # we assume the pattern is binary (e.g. 0 or 1)
        self.ledpattern = np.zeros((self.NLeds, 3))
        
        # this is a istance of the _parent32Client class
        self._parent = parent
        self.timeout = 1
        self.intensity = (255,255,255)
        
        self.ledArrayMode = {
            "array": 0,
            "full": 1,
            "single": 2,
            "off": 3,
            "left": 4,
            "right": 5,
            "top": 6,
            "bottom": 7,
            "multi": 8}
            
            
    
    '''
    ##############################################################################################################################
    LED ARRAY
    ##############################################################################################################################
    '''
    def send_LEDMatrix_array(self, led_pattern, is_blocking = False, timeout=gTimeout):
        '''
        Send an LED array pattern e.g. an RGB Matrix: led_pattern=np.zeros((3,8,8))
        '''
        path = '/ledarr_act'
        # if we have a 2d pattern => flatten
        if len(led_pattern.shape)==3:
            led_pattern=np.reshape(led_pattern, (np.prod(led_pattern.shape[0:2]),led_pattern.shape[2])) 

        # convert pattern strip to list of RGB values
        pattern_list = []
        for i in range(led_pattern.shape[0]):
            pattern_list.append({
                "id": i,
                "r": int(led_pattern[i,0]),
                "g": int(led_pattern[i,1]),
                "b": int(led_pattern[i,2])                        
            })

        payload = {
            "task":path,
            "led": {
                "LEDArrMode": self.ledArrayMode["array"],
                "led_array": pattern_list
            }
        }
        #self._parent.logger.debug("Setting LED Pattern (array) ")
        r = self._parent.post_json(path, payload, getReturn=is_blocking, timeout=timeout)
        
        return r

    def send_LEDMatrix_full(self, intensity = (255,255,255), getReturn=True, timeout=gTimeout):
        '''
        set all LEDs with te same RGB value: intensity=(255,255,255)
        '''
        #{"task": "/ledarr_act","led": {"LEDArrMode": 1,"led_array": [{"b": 255,"g": 255,"r": 255}]}}
        path = '/ledarr_act'
        payload = {
            "task":path,
            "led": {
                "led_array":[{
                    "r": int(intensity[0]),
                    "g": int(intensity[1]),
                    "b": int(intensity[2])}],
                "LEDArrMode": self.ledArrayMode["full"]
            }
        }
        
        #self._parent.logger.debug("Setting LED Pattern (full): "+ str(intensity))
        r = self._parent.post_json(path, payload, getReturn=getReturn, timeout=timeout)
        return r

    def send_LEDMatrix_special(self, pattern="left", intensity = (255,255,255),timeout=gTimeout):
        '''
        set all LEDs inside a certain pattern (e.g. left half) with the same RGB value: intensity=(255,255,255), rest 0
        '''
        path = '/ledarr_act'
        payload = {
            "task":path,
            "led": {
            "r": int(intensity[0]),
            "g": int(intensity[1]),
            "b": int(intensity[2]),
            "LEDArrMode": self.ledArrayMode[pattern]
            }
        }
        self._parent.logger.debug("Setting LED Pattern (full): "+ str(intensity))
        r = self._parent.post_json(path, payload, getReturn=True, timeout=timeout)
        return r

    def send_LEDMatrix_single(self, indexled=0, intensity=(255,255,255), timeout=gTimeout):
        '''
        update only a single LED with a colour:  indexled=0, intensity=(255,255,255)
        '''
        path = '/ledarr_act'
        payload = {
            "task":path,
            "led": {
                "LEDArrMode": self.ledArrayMode["single"],
                "led_array": [{
                    "id": indexled,
                    "r": int(intensity[0]),
                    "g": int(intensity[1]),
                    "b": int(intensity[2])}]
            }
        }
        self._parent.logger.debug("Setting LED Pattern (single) ")
        r = self._parent.post_json(path, payload, getReturn=True, timeout=timeout)
        return r


    def get_LEDMatrix(self, timeout=gTimeout):
        '''
        get information about pinnumber and number of leds
        '''
        # TOOD: Not implemented yet
        path = "/ledarr_get"
        payload = {
            "task":path
        }
        r = self._parent.post_json(path, payload, getReturn=True, timeout=timeout)
        return r

    def setSingle(self, indexled, state):
        # Update the intensity of a single LED in a cartesian grid        
        ix = indexled//self.Nx
        iy = indexled%self.Nx
        self.ledpattern[indexled] = (state,state,state) # either [0, 1]
        # forward backward enumaration
        if ix%2 != 0:
           indexled = (ix*self.Nx)+(self.Ny-iy-1)
        self.send_LEDMatrix_single(indexled=indexled, intensity=np.array(state)*np.array(self.intensity), timeout=self.timeout)

        return self.ledpattern
    
    def setAll(self, state, intensity=None):
        # fast addressing
        # turns on all LEDs at a certain intensity
        if intensity is not None:
            self.intensity = intensity
        intensity2display = np.array(self.intensity)*np.array(state)
        self.send_LEDMatrix_full(intensity = intensity2display, timeout=self.timeout)
        self.ledpattern = state*np.ones((self.NLeds, 3))
        return self.ledpattern
    
    def setIntensity(self, intensity):
        self.intensity = intensity
        self.setPattern()
    
    def setPattern(self, ledpattern=None):
        # sends pattern with proper intensity
        if ledpattern is not None:
            self.ledpattern = ledpattern
        pattern2send = (self.ledpattern>=1)*self.intensity
        self.send_LEDMatrix_array(pattern2send, timeout=self.timeout)
        return self.ledpattern
    
    def getPattern(self):
        return self.ledpattern
        
    def set_led(self, colour=(0,0,0)):
        payload = {
            "r": colour[0],
            "g": colour[1],
            "b": colour[2]
        }
        path = '/led'
        r = self._parent.post_json(path, payload)
        return r

    def set_ledpin(self, ledArrPin=4, ledArrNum=None):
        if ledArrNum is None:
            ledArrNum = self.NLeds
            
        path = "/ledarr_set"
        payload = {
            "task": path,
            "ledArrPin": ledArrPin,
            "ledArrNum": ledArrNum
        }
        r = self._parent.post_json(path, payload, getReturn=True)
        return r
        
    def get_ledpin(self):
        path = "/ledarr_get"
        payload = {
            "task": path,
        }
        r = self._parent.post_json(path, payload, getReturn=True, timeout=1)
        return r
            
    
    def setLEDArrayConfig(self, ledArrPin=4, ledArrNum=None):
        # make imswitch happy
        return self.set_ledpin(ledArrPin=ledArrPin, ledArrNum=ledArrNum)


import time
import json        
        
class Home(object):
    ## Laser
    def __init__(self, parent):
        self._parent = parent
        self.endposrelease = 3000
        self.direction = 1
        self.speed = 15000
        self.timeout = 20000
        
    def home_x(self, speed = None, direction = None, endposrelease = None, timeout=None, isBlocking=False):
        # axis = 1 corresponds to 'X'
        axis = 1
        self.home(axis=axis, 
                  timeout=timeout, 
                  speed = speed, 
                  direction = direction, 
                  endposrelease=endposrelease,
                  isBlocking=isBlocking)

    def home_y(self, speed = None, direction = None, endposrelease = None, timeout=None, isBlocking=False):
        # axis = 2 corresponds to 'Y'
        axis = 2
        self.home(axis=axis, 
                  timeout=timeout, 
                  speed = speed, 
                  direction = direction, 
                  endposrelease=endposrelease, 
                  isBlocking=isBlocking)    
    
    def home_z(self, speed = None, direction = None, endposrelease = None, timeout=None, isBlocking=False):
        # axisa = 3 corresponds to 'Z'
        axis = 3
        self.home(axis=axis, 
                  timeout=timeout, 
                  speed = speed, 
                  direction = direction, 
                  endposrelease=endposrelease, 
                  isBlocking=isBlocking)    
        
    def home(self, axis=None, timeout=None, speed=None, direction=None, endposrelease=None, isBlocking=False):
        '''
        axis = 0,1,2,3 or 'A, 'X','Y','Z'
        timeout => when to stop homing (it's a while loop on the MCU)
        speed => speed of homing (0...15000)
        direction => 1,-1 (left/right)
        endposrelease => how far to move after homing (0...3000)
        '''
        
        # default values
        if speed is None:
            speed = self.speed
        if direction is None:
            direction = self.direction
        if endposrelease is None:
            endposrelease = self.endposrelease
        if timeout is None:
            timeout = self.timeout

        if direction not in [-1,1]:
            direction = 1

        # construct json string
        path = "/home_act"

        payload = {
            "task": path,
            "home":{
                "steppers": [
                {
                 "stepperid": axis,
                 "timeout":timeout*1000,
                 "speed":speed,
                 "direction":direction,
                 "endposrelease":endposrelease
                 }]
            }}
     
        # send json string
        r = self._parent.post_json(path, payload, getReturn=True, timeout=timeout)
        
        # wait until job has been done        
        time0=time.time()
        if isBlocking and self._parent.serial.is_connected:
            while True:
                time.sleep(0.05) # don'T overwhelm the CPU
                # see if already done
                try:
                    rMessage = self._parent.serial.serialdevice.readline().decode() # TODO: Make sure it's compatible with all motors running at the same time
                except Exception as e:
                    self._parent.logger.error(e)
                    rMessage = ""
                # check if message contains a motor that is done already
                if rMessage.find('isDone') >-1:
                    break
                if time.time()-time0>timeout:
                    break


        
        
        
        return r




from PIL import Image
import base64
import io
import numpy as np

class Camera(object):
    
    def __init__(self, parent, width=320, height=240, fps=10):
        self._parent = parent
        self.width = width
        self.height = height
        self.lastFrame = 255*np.random.randn(self.height, self.width)
        
        
    def set_camera(self, width=None, height=None, fps=None, timeout=3):
        path = "/camera_set"
        payload = {"path":path,
                   "width":width,
                   "height":height,
                   "fps":fps}
        r = self._parent.post_json(path, payload, timeout=timeout)
        return r
    
    def get_camera(self):
        path = "/camera_get"
        return None
    
    def get_frame(self, timeout=5):
        path = "/camera_act"
        payload = {
            "grabimage":1
        }
        r = self._parent.post_json(path, payload, timeout=timeout)
        
        #%%

        try:
            #read image and decode
            imageB64 = r['frame']
            image = np.array(Image.open(io.BytesIO(base64.b64decode(imageB64))))

        except Exception as e:
            self._parent.logger.error(f"Error: {e}")
            image = None
            image = self.lastFrame
            
        return image
            

    def set_led(self, value):
        path = "/led_act"
        payload = {
            "value":value
        }
        r = self._parent.post_json(path, payload, timeout=2)
        
        return r



class SLM(object):
    
    
    def __init(self, parent):
        self._parent = parent
        
    '''
    ##############################################################################################################################
    SLM
    ##############################################################################################################################
    '''
    def send_SLM_circle(self, posX, posY, radius, color, timeout=1):
        '''
        Send an LED array pattern e.g. an RGB Matrix: led_pattern=np.zeros((3,8,8))
        '''
        path = '/slm_act'
        payload = {
            "posX": posX,
            "posY": posY,
            "radius": radius,
            "color": color,
            "slmMode": "circle"
        }
        r = self.post_json(path, payload, timeout=timeout)
        return r

    def send_SLM_clear(self, timeout=1):
        '''
        Send an LED array pattern e.g. an RGB Matrix: led_pattern=np.zeros((3,8,8))
        '''
        path = '/slm_act'
        payload = {
            "slmMode": "clear"
        }
        r = self.post_json(path, payload, timeout=timeout)
        return r

    def send_SLM_full(self, color, timeout=1):
        '''
        Send an LED array pattern e.g. an RGB Matrix: led_pattern=np.zeros((3,8,8))
        '''
        path = '/slm_act'
        payload = {
            "color":color,
            "slmMode": "full"
        }
        r = self.post_json(path, payload, timeout=timeout)
        return r


    def send_SLM_image(self, image, startX, startY, timeout=1):
        '''
        Send an LED array pattern e.g. an RGB Matrix: led_pattern=np.zeros((3,8,8))
        '''
        path = '/slm_act'

        endX = startX+image.shape[0]
        endY = startY+image.shape[1]

        payload = {
            "color": image[:].flatten().tolist(),
            "startX":startX,
            "startY":startY,
            "endX":endX,
            "endY":endY,
            "slmMode": "image"
        }
        r = self.post_json(path, payload, timeout=timeout)
        return r



import subprocess
import os
import esptool
import tempfile
import urllib
import threading
import zipfile
import progressbar 
import platform
import glob
try:
    import requests
    is_requests = True
except:
    is_requests = False

class updater(object):
    
    def __init__(self, ESP32=None, port=None, parent=None):
        if ESP32 is not None:
            self.port = ESP32.serial.serialport
        if port is not None:
            self.port = port
        # parent holds the entire esp object
        if parent is not None:
            self._parent = parent
            self.port = self._parent.serial.serialport

        # define a temporary firmware file name for the firmware download
        self.firmwarePath = os.path.join(tempfile.gettempdir(), "uc2rest")
            

                
    def unzipFiles(self):
        ''' Unzip the UC2Rest.zip '''
        print("Unzipping files")
        with zipfile.ZipFile(os.path.join(self.firmwarePath, self.uc2restZip), 'r') as zip_ref:
            zip_ref.extractall(self.firmwarePath)
        # list the files in the directory
        filenames = os.listdir(self.firmwarePath)
        print("Done unzipping files:")
        print(filenames)
        return filenames
        
        
    def flashFirmware(self, firmwarePath=None):
        # sideload the firmware if already available online
        
        self.filenames = self.unzipFiles()
        if self._parent is not None:
            # in case the serial is still open, perhaps it makes sense to close it
            try:
                self._parent.serial.closeSerial()
            except:
                pass
            
        if platform.system()=="Windows":
            try:
                #esptool.py --chip esp32 --port /dev/cu.SLAB_USBtoUART --baud 921600 
                # --before default_reset --after hard_reset write_flash -z --flash_mode dio 
                # --flash_freq 80m --flash_size 4MB 
                # 0x1000 ./ESP32/build/main.ino.bootloader.bin 
                # 0x8000 ./ESP32/build/main.ino.partitions.bin 
                # 0xe000 ./ESP32/build/boot_app0.bin 
                # 0x10000 ./ESP32/build/main.ino.bin 
                try:
                    cmd = ["esptool.py", 
                            "--chip", "esp32",
                            "--port", self.port, 
                            "--baud", "921600", 
                            "write_flash",
                            "--flash_freq", "80m",
                            "--flash_mode", "dio", 
                            "--flash_size", "detect", 
                            "0xe000", os.path.join(self.firmwarePath,"boot_app0.bin"),
                            "0x1000", os.path.join(self.firmwarePath,"main.ino.bootloader.bin"), 
                            "0x8000", os.path.join(self.firmwarePath,"main.ino.partitions.bin"), 
                            "0x10000", os.path.join(self.firmwarePath,"main.ino.bin")]
                    print('Using command %s' % ' '.join(cmd))
                    process = subprocess.Popen(cmd,shell=True,stderr=subprocess.PIPE, stdout=subprocess.PIPE)
                    process.wait()
                    stdout, stderr = process.communicate()
                    if str(stderr).find("not reco")>0 or not str(stdout).find("Leaving...")>0: 
                        raise Exception
                    
                except Exception as e:
                    print(e)
                    print("We will try an alternative route:")
                    try:
                        cmd = ["python -m esptool", 
                                "--chip", "esp32",
                                "--port", self.port, 
                                "--baud", "921600", 
                                "write_flash",
                                "--flash_freq", "80m",
                                "--flash_mode", "dio", 
                                "--flash_size", "detect", 
                                "0xe000", os.path.join(self.firmwarePath,"boot_app0.bin"),
                                "0x1000", os.path.join(self.firmwarePath,"main.ino.bootloader.bin"), 
                                "0x8000", os.path.join(self.firmwarePath,"main.ino.partitions.bin"), 
                                "0x10000", os.path.join(self.firmwarePath,"main.ino.bin")]
                        print('Using command %s' % ' '.join(cmd))
                        env = os.environ
                        process = subprocess.Popen(cmd,shell=True, env=env,stderr=subprocess.PIPE, stdout=subprocess.PIPE)
                        process.wait()
                        stdout, stderr = process.communicate()
                        if str(stderr).find("not reco")>0 or  not str(stdout).find("Leaving...")>0: 
                            raise Exception
                    except Exception as e:
                        print(e)
                        print("We will try an alternative route:")
                        try:
                            # special weindows case?
                            cmd = ["esptool", 
                                    "--chip", "esp32",
                                    "--port", self.port, 
                                    "--baud", "921600", 
                                    "write_flash",
                                    "--flash_freq", "80m",
                                    "--flash_mode", "dio", 
                                    "--flash_size", "detect", 
                                    "0xe000", os.path.join(self.firmwarePath,"boot_app0.bin"),
                                    "0x1000", os.path.join(self.firmwarePath,"main.ino.bootloader.bin"), 
                                    "0x8000", os.path.join(self.firmwarePath,"main.ino.partitions.bin"), 
                                    "0x10000", os.path.join(self.firmwarePath,"main.ino.bin")]
                            print('Using command %s' % ' '.join(cmd))
                            env = os.environ
                            process = subprocess.Popen(cmd,shell=True, env=env)
                            process.wait()
                        except Exception as e:
                            print(e)
                            print("Firmware not flashed.")
                            return False                    
                print("Firmware flashed")
                return True
            except Exception as e:
                print(e)
                print("Firmware flash failed")
                return False
            
        else:
            try:
                try:
                    cmd = ["esptool.py", 
                            "--chip", "esp32",
                            "--port", self.port, 
                            "--baud", "921600", 
                            "write_flash",
                            "--flash_freq", "80m",
                            "--flash_mode", "dio", 
                            "--flash_size", "detect", 
                            "0xe000", os.path.join(self.firmwarePath,"boot_app0.bin"),
                            "0x1000", os.path.join(self.firmwarePath,"main.ino.bootloader.bin"), 
                            "0x8000", os.path.join(self.firmwarePath,"main.ino.partitions.bin"), 
                            "0x10000", os.path.join(self.firmwarePath,"main.ino.bin")]
                    print('Using command %s' % ' '.join(cmd))
                    process = subprocess.Popen(cmd)
                    process.wait()
                except Exception as e:
                    print(e)
                    print("We will try an alternative route:")
                    cmd = ["python -m esptool", 
                            "--chip", "esp32",
                            "--port", self.port, 
                            "--baud", "921600", 
                            "write_flash",
                            "--flash_freq", "80m",
                            "--flash_mode", "dio", 
                            "--flash_size", "detect", 
                            "0xe000", os.path.join(self.firmwarePath,"boot_app0.bin"),
                            "0x1000", os.path.join(self.firmwarePath,"main.ino.bootloader.bin"), 
                            "0x8000", os.path.join(self.firmwarePath,"main.ino.partitions.bin"), 
                            "0x10000", os.path.join(self.firmwarePath,"main.ino.bin")]
                    print('Using command %s' % ' '.join(cmd))
                    process = subprocess.Popen(cmd)
                    process.wait()
                print("Firmware flashed")
                return True
            except Exception as e:
                print(e)
                print("Firmware flash failed")
                return False

    
    def downloadFirmware(self):

        print("We are checking for pre-built binaries on Github")
        self.firmwareDownloadPath = 'https://api.github.com/repos/youseetoo/uc2-esp32/releases/latest'
        releaseResponse = requests.get(
            self.firmwareDownloadPath
        )
        latestVersion = releaseResponse.json()['tag_name']
        print("Latest version is: "+latestVersion)

        ## attempting to downloda the current ImSwitch version
        self.downloadURL = releaseResponse.json()['assets'][0]['browser_download_url']
        self.uc2restZip = "UC2Rest.zip"
        print("We are downloading the software from: "+self.uc2restZip)
                

        print("Downloading Firmware from "+self.firmwareDownloadPath)
        # download the firmware from github
        ## inplace replacement won't work I guess? => seems to work
        def dlImSwitch(downloadURL, fileName):
            resultDL = urllib.request.urlretrieve(downloadURL, fileName, MyProgressBar())
            print("Done Downloading")


        try:
            if not os.path.exists(self.firmwarePath):
                os.makedirs(self.firmwarePath)
            # remove file if exists
            if os.path.exists(self.firmwarePath+self.uc2restZip):
                os.remove(os.path.join(self.firmwarePath,self.uc2restZip))
            
            # download the new version in a separate thread
            mThread =  threading.Thread(target=dlImSwitch, args=(self.downloadURL, os.path.join(self.firmwarePath,self.uc2restZip)))
            mThread.start()
            mThread.join()

            print("Succesfully downloaded file: "+os.path.join(self.firmwarePath, self.uc2restZip))
            return True
        except Exception as e:
            print(e)
            print("Firmware download failed"+self.uc2restZip)

            return False
            
    def removeFirmware(self):
        try:
            print("Removing Firmware:"+self.uc2restZip)
            for ifile in glob.glob(self.firmwarePath+"/*"):
                try:os.remove(ifile)
                except:pass
            return True
        except Exception as e:
            print(e)
            return False
        
        

class MyProgressBar():
    def __init__(self):
        self.pbar = None

    def __call__(self, block_num, block_size, total_size):
        if not self.pbar:
            self.pbar=progressbar.ProgressBar(maxval=total_size)
            self.pbar.start()

        downloaded = block_num * block_size
        if downloaded < total_size:
            self.pbar.update(downloaded)
        else:
            self.pbar.finish()
    
        
if __name__ == "__main__":
    updater = updater(port="/dev/cu.SLAB_USBtoUART", firmwarePath=None)
    updater.downloadFirmware()
    updater.flashFirmware()
    
    # remove firmware.bin after flashing
    updater.removeFirmware()
    
            
        
                      